/* MAKITranslator.cpp */
/* MAKI/MAG=>BITMAP(32bit) */

#include	<Be.h>
#include	<TranslatorAddOn.h>

#define	VERSION	"ver.0.02     1999.4.11"
#define	BUFSIZE	0x10000

char	translatorName[]="MAKI Translator";	
char	translatorInfo[]="MAKI file format translator "VERSION;
int32	translatorVersion=000;

const char	rmode_name[2][16]={	"normal","correct aspect"	};
const char	wmode_name[2][16]={	"16colors","256colors"	};
const char	mag_dx[16]={	 0,-1,-2,-4, 0,-1, 0,-1,-2, 0,-1,-2, 0,-1,-2, 0		};
const char	mag_dy[16]={	 0, 0, 0, 0,-1,-1,-2,-2,-2,-4,-4,-4,-8,-8,-8,-16	};

static char	rmode=1;
static char	wmode=1;

translation_format	inputFormats[]={
	{	B_TRANSLATOR_BITMAP,	B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/x-be-bitmap",	"Be Bitmap Format"		},
	{	'MAKI',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/maki",			"MAKI bitmap format"	},
	{	'MAG ',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/mag",			"MAG bitmap format"		},
	{0,0,0,0,{""},{""}}
};
translation_format	outputFormats[]={
	{	B_TRANSLATOR_BITMAP,	B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/x-be-bitmap",	"Be Bitmap Format"		},
//	{	'MAKI',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/maki",			"MAKI bitmap format"	},
	{	'MAG ',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/mag",			"MAG bitmap format"		},
	{0,0,0,0,{""},{""}}
};

struct MAKIChecker{		//MAKI Checker format(size=32)
	char		mark[8];		//	"MAKI01A","MAKI01B","MAKI02  "
	char		hard[4];		//  hardware
	char		user[20];		//  username
};

struct MAKIHeader{		//MAKI Header format(size=16)
	uint16		fb_size;		//	flag_b size
	uint16		pa_size;		//	pixel_a size
	uint16		pb_size;		//	pixel_b size
	uint16		exflag;
	uint16		xmin,ymin,xmax,ymax;	//(xmin,ymin)-(xmax-1,ymax-1)
};

struct MAGHeader{		//MAG Header format(size=32)
	char		top;			//	top mark
	char		hardc;			//  hardware code
	char		hardf;			//  hardware flag
	char		scrn;			//  screen mode
	uint16		xmin,ymin,xmax,ymax;
	uint32		fa_off;			//	flag_a offset
	uint32		fb_off;			//	flag_b offset
	uint32		fb_size;		//	flag_b size
	uint32		p_off;			//	pixel offset
	uint32		p_size;			//	pixel size
};

uchar	*rb,*wb;				//buffer
//-------------------------------------------------------------------------------------
class MAKIWindow:public BWindow
{
	public:
		MAKIWindow(BRect area,const char *title,window_type kind,uint32 flags);
		bool	QuitRequested();
};

class configview:public BView
{
	public:
		configview(BRect frame);
		void	AttachedToWindow();
		void	MessageReceived(BMessage *msg);
	private:
		BRadioButton	*rb1[5],*rb2[4];
};

MAKIWindow::MAKIWindow(BRect area,const char *title,window_type kind,uint32 flags):BWindow(area,title,kind,flags)
{
}

bool MAKIWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return(true);
}

configview::configview(BRect frame):BView(frame,"MAKI_Translator Settings",B_NOT_RESIZABLE,B_WILL_DRAW)
{
	BStringView	*sv;
	BMessage	*mes;
	BBox		*box;
	sv=new BStringView(BRect(10,3,190,20),"title","MAKITranslator  (MAKI/MAG)");
	sv->SetFont(be_bold_font);
	AddChild(sv);
	sv=new BStringView(BRect(40,20,180,37),"ver",VERSION);
	sv->SetFont(be_bold_font);
	AddChild(sv);
	sv=new BStringView(BRect(10,113,190,130),"program","code:  (c)Hiroyuki Tsutsumi");
	sv->SetFont(be_bold_font);
	AddChild(sv);
	sv=new BStringView(BRect(10,130,190,147),"format","format:  (c)Woody-RINN");
	sv->SetFont(be_bold_font);
	AddChild(sv);
	box=new BBox(BRect(0,48,100,102),"box1");
	box->SetLabel("read");
	for(int i=0;i<2;i++){
		mes=new BMessage('rmod');mes->AddInt8("aspect_mode",i);
		rb1[i]=new BRadioButton(BRect(5,i*15+15,95,i*15+30),"radiobutton",rmode_name[i],mes);
		box->AddChild(rb1[i]);
	}
	AddChild(box);
	box=new BBox(BRect(100,48,200,102),"box2");
	box->SetLabel("write(MAG)");
	for(int i=0;i<2;i++){
		mes=new BMessage('wmod');mes->AddInt8("encode_mode",i);
		rb2[i]=new BRadioButton(BRect(5,i*15+15,95,i*15+30),"radiobutton",wmode_name[i],mes);
		box->AddChild(rb2[i]);
	}
	AddChild(box);
}

void configview::AttachedToWindow()
{
	SetViewColor(233,255,233);
	SetLowColor(122,155,122);
	for(int i=0;i<2;i++){
		rb1[i]->SetTarget(this);
		if(rmode==i)rb1[i]->SetValue(1);
	}
	for(int i=0;i<2;i++){
		rb2[i]->SetTarget(this);
		if(wmode==i)rb2[i]->SetValue(1);
	}
}

void configview::MessageReceived(BMessage *msg)
{
	switch(msg->what){
		case 'rmod':{
			beep();
			msg->FindInt8("aspect_mode",(signed char*)&rmode);
		}break;
		case 'wmod':{
			beep();
			msg->FindInt8("encode_mode",(signed char*)&wmode);
		}break;
		default:	BView::MessageReceived(msg);
	}
}
//-------------------------------------------------------------------------------------
main()
{
	BApplication app("application/x-vnd.Beer-MAKI_Translator");
	MAKIWindow *window=new MAKIWindow(BRect(100,100,400,300),"MAKI Settings",B_TITLED_WINDOW,B_NOT_RESIZABLE|B_NOT_ZOOMABLE);
	BView *config=NULL;
	BRect extent;
	MakeConfig(NULL,&config,&extent);
	window->ResizeTo(extent.Width(),extent.Height());
	window->AddChild(config);
	window->Show();
	app.Run();
}

status_t Identify(BPositionIO *inSource,const translation_format *inFormat,BMessage *ioExtension,translator_info *outInfo,uint32 outType)
{
	if((outType)&&(outType!=B_TRANSLATOR_BITMAP)&&(outType!='MAG '))return(B_NO_TRANSLATOR);
	const int			bufsize=64;
	char				buf[bufsize];
	TranslatorBitmap	*bmphd=(TranslatorBitmap*)buf;
	MAKIChecker			*makick=(MAKIChecker*)buf;
	MAKIHeader			*makihd=(MAKIHeader*)(buf+32);
	MAGHeader			*maghd=(MAGHeader*)(buf+32);

	memset(buf,0,bufsize);
	//	read checker(MAKI/MAG)/header(TranslatorBitmap)
	ssize_t		len,stt;
	stt=inSource->Read(buf,len=32);
	if(stt<len)return(stt);

	//	check input-type
	int		itype=-1;
	char	itype2='?';
	if(B_BENDIAN_TO_HOST_INT32(bmphd->magic)==B_TRANSLATOR_BITMAP){
		itype=0;						//B_TRANSLATOR_BITMAP
	}else if(!strncmp(makick->mark,"MAKI01A ",8)){
		itype=1;itype2='a';				//MAKI-A
	}else if(!strncmp(makick->mark,"MAKI01B ",8)){
		itype=1;itype2='b';				//MAKI-B
	}else if(!strncmp(makick->mark,"MAKI02  ",8)){
		itype=2;						//MAG
	}else return(B_NO_TRANSLATOR);

	//	read header(MAKI/MAG)
	if(itype==1){
		len=sizeof(MAKIHeader);
		stt=inSource->Read(makihd,len);
		if(stt<len)return(stt);
	}else if(itype==2){
		if(makick->user[19]!=0x1a)while(1){
			stt=inSource->Read(maghd,1);
			if(stt<1)return(stt);
			if(maghd->top==0x1a)break;
		}
		len=sizeof(MAGHeader);
		stt=inSource->Read(maghd,len);
		if(stt<len)return(stt);
	}

	//	check header & out info
	if(itype==0){						//B_TRANSLATOR_BITMAP
		if(B_BENDIAN_TO_HOST_INT32(bmphd->colors)!=B_RGB_32_BIT)return(B_NO_TRANSLATOR);
		sprintf(outInfo->name,"%s, type %d",inputFormats[2].name,B_BENDIAN_TO_HOST_INT32(bmphd->colors));
	}else if(itype==1){					//MAKI
		sprintf(outInfo->name,"%s, type %c (%d colors)",inputFormats[0].name,itype2,16);
	}else{								//MAG
		sprintf(outInfo->name,"%s, (%d colors)",inputFormats[1].name,maghd->scrn&0x80?256:16);
	}
	outInfo->type=inputFormats[itype].type;
	outInfo->translator=0;
	outInfo->group=inputFormats[itype].group;
	outInfo->quality=inputFormats[itype].quality;
	outInfo->capability=inputFormats[itype].capability;
	strcpy(outInfo->MIME,inputFormats[itype].MIME);
	return(B_OK);
}

static status_t CopyLoop(BPositionIO *input,BPositionIO *output)
{																//BITMAP->BITMAP , MAKI->MAKI
	rb=new uchar[BUFSIZE];			if(!rb)return(B_NO_MEMORY);
	ssize_t		len,stt;
	while((stt=input->Read(rb,len=BUFSIZE))>0)if((stt=output->Write(rb,len=stt))<0)break;
	return(stt);
}

static status_t WriteMAKI(BPositionIO *input,BPositionIO *output,BMessage *ioExtension)
{																//BITMAP->MAKI
	MAKIChecker			makick;
	MAGHeader			maghd;
	TranslatorBitmap	bmphd;
	uint32				pal[256];
	char		mode;
	ssize_t		len,stt;

	//	read bitmap header
	stt=input->Read(&bmphd,len=sizeof(bmphd));
	if(stt<len)return(stt);
	bmphd.magic=B_BENDIAN_TO_HOST_INT32(bmphd.magic);
	bmphd.bounds.left=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.left);
	bmphd.bounds.top=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.top);
	bmphd.bounds.right=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.right);
	bmphd.bounds.bottom=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.bottom);
//	bmphd.colors=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
	*(int32*)(&(bmphd.colors))=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
	bmphd.rowBytes=B_BENDIAN_TO_HOST_INT32(bmphd.rowBytes);
	bmphd.dataSize=B_BENDIAN_TO_HOST_INT32(bmphd.dataSize);
	if(bmphd.colors!=B_RGB_32_BIT)return(B_NO_TRANSLATOR);

	maghd.xmin=(int)bmphd.bounds.left&0xfff8;maghd.xmax=(int)bmphd.bounds.right|0x0007;//for odd_pixels
	maghd.ymin=bmphd.bounds.top;maghd.ymax=bmphd.bounds.bottom;
	int	w=maghd.xmax-maghd.xmin+1;
	int	h=maghd.ymax-maghd.ymin+1;
	uint	rbsize=w*h*4;
	rb=new uchar[rbsize];		if(!rb)return(B_NO_MEMORY);
	memset(rb,0,rbsize);					//clr input-area
	uint32	*fgp,*fg=(uint32*)rb;
	stt=input->Read(rb,len=bmphd.dataSize);		//read all
	if(stt<len)return(stt);

	//count colors
	memset(pal,0,4*256);
	int		cols=(wmode==0)?16:256;
	bool	pmode=true;
	int		cnt=0;
	fgp=fg;
	for(int z=0;z<w*h;z++){
		uint32	c=B_BENDIAN_TO_HOST_INT32(*(fgp++))&0xffffff00;
		int	i;
		for(i=0;i<cnt;i++)if(c==pal[i])break;
		if(i==cnt){
			if(cnt==cols){pmode=false;break;}
			pal[cnt++]=c;
		}
	}
	if(pmode){
		cols=(cnt<=16)?16:256;
	}else{
		int	i=0;
		if(cols==16){					//8colors
			for(int b=0;b<2;b++)for(int g=0;g<2;g++)for(int r=0;r<2;r++)pal[i++]=((b<<24)|(g<<16)|(r<<8))*255;
		}else{							//216colors
			for(int b=0;b<6;b++)for(int g=0;g<6;g++)for(int r=0;r<6;r++)pal[i++]=((b<<24)|(g<<16)|(r<<8))*51;
		}
		for(;i<cols;i++)pal[i]=0;
	}
	if(cols>16){																	//for odd_pixels
		maghd.xmin=(int)bmphd.bounds.left&0xfffc;maghd.xmax=(int)bmphd.bounds.right|0x0003;
		w=maghd.xmax-maghd.xmin+1;
	}
//printf("cnt=%d\n",cnt);
//if(pmode)printf("palette mode\n");
//printf("cols=%d\n",cols);
	uint	fasize,fbsize,fcsize,ffsize;
	ffsize=(cols>16)?w*h:w*h/2;
	uchar	*fap,*fa=rb;				//flaga(write)
	uchar	*fbp,*fb=rb+rbsize/8;		//flagb(write)
	uchar	*fcp,*fc=rb+rbsize/4;		//pixel(write)
	uchar	*fdp,*fd=rb+rbsize/2;		//copy-position		fdsize=w*h/4,2
	uchar	*fep,*fe=rb+rbsize*5/8;		//work
	uchar	*ffp,*ff=rb+rbsize-ffsize;	//color

	//	xfer data
	int	xd=(cols>16)?2:4;
	ffp=ff+ffsize;fgp=(uint32*)(rb+bmphd.dataSize);
	for(int y=h-1;y>=0;y--)for(int x=maghd.xmax;x>=maghd.xmin;x--){
		uchar	p=0;
		if((x>=bmphd.bounds.left)&&(x<=bmphd.bounds.right)){
			uint32	c=B_BENDIAN_TO_HOST_INT32(*(--fgp))&0xffffff00;
			if(pmode){
				while(c!=pal[p])p++;
			}else if(cols==16){			//8colors
				p=((c>>29)&0x04)|((c>>22)&0x02)|((c>>15)&0x01);
			}else{						//216colors
				p=(int)(((c>>24)&0xff)/43)*36+(int)(((c>>16)&0xff)/43)*6+(int)(((c>>8)&0xff)/43);
			}
		}
		if((cols>16)||(x&1)){
			*(--ffp)=p;
		}else{
			(*ffp)|=p<<4;
		}
	}
	memset(rb,0,rbsize-ffsize);
	fcp=fc;fdp=fd;ffp=ff;
	for(int y=0;y<h;y++)for(int x=0;x<w/xd;x++){
		int		xx,yy,d,i;
		uint16	p=*(uint16*)ffp;
		ffp+=2;
		for(i=0;i<16;i++){
			d=i;
			if((d==0)&&(y>0))d=*(fdp-w/xd);
			if(d==0)continue;
			if(((xx=x+mag_dx[d])>=0)&&((yy=y+mag_dy[d])>=0)){
				if(*(uint16*)(ff+(w/xd*yy+xx)*2)==p)break;
			}
		}
		if(i>15)d=0;
		*(fdp++)=d;
		if(d==0){*(uint16*)fcp=p;fcp+=2;}
	}
	fcsize=fcp-fc;
	fap=fa-1;fbp=fb;fdp=fd;
	uchar	b=0;
	for(int y=0;y<h;y++){
		fep=fe;
		for(int x=0;x<w/xd/2;x++){
			(*(uint16*)fep)=(*(uint16*)fdp)^*(uint16*)(fdp-w/xd);
//			(*fep)=(*fdp)^*(fdp-w/xd);*(fep+1)=*(fdp+1)^*(fdp-w/xd+1);
			uchar	d;
			d=(*fep)<<4;
			d|=*(fep+1);
			if(b>1){
				b>>=1;
			}else{
				b=0x80;
				fap++;
			}
			if(d){
				*(fbp++)=d;
				(*fap)|=b;
			}
			fdp+=2;fep+=2;
		}
	}
	if(b<=1)fap++;
	fasize=fap-fa;
	if(fasize&1)fasize++;
	fbsize=fbp-fb;
	if(fbsize&1)fbsize++;
	//	write MAG header
	strncpy(makick.mark,"MAKI02  ",8);
	strncpy(makick.hard,"BeOS",4);
	strncpy(makick.user,"MAKITranslator v002",19);
	makick.user[19]=0x1a;
	stt=output->Write(&makick,len=sizeof(makick));
	if(stt<len)return(stt);
	maghd.top=0x00;
	maghd.hardc=0x00;
	maghd.hardf=0x00;
	maghd.scrn=(cols>16)?0x80:0x00;
	maghd.xmin=B_HOST_TO_LENDIAN_INT16(bmphd.bounds.left);
	maghd.ymin=B_HOST_TO_LENDIAN_INT16(bmphd.bounds.top);
	maghd.xmax=B_HOST_TO_LENDIAN_INT16(bmphd.bounds.right);
	maghd.ymax=B_HOST_TO_LENDIAN_INT16(bmphd.bounds.bottom);
	maghd.fa_off=B_HOST_TO_LENDIAN_INT32(32+3*cols);				//	flag_a offset
	maghd.fb_off=B_HOST_TO_LENDIAN_INT32(32+3*cols+fasize);			//	flag_b offset
	maghd.fb_size=B_HOST_TO_LENDIAN_INT32(fbsize);					//	flag_b size
	maghd.p_off=B_HOST_TO_LENDIAN_INT32(32+3*cols+fasize+fbsize);	//	pixel offset
	maghd.p_size=B_HOST_TO_LENDIAN_INT32(fcsize);					//	pixel size
	stt=output->Write(&maghd,len=sizeof(maghd));
	if(stt<len)return(stt);
	stt=output->SetSize(sizeof(makick)+sizeof(maghd)+fasize+fbsize+fcsize);
	if(stt<B_OK)return(stt);

	for(int i=0;i<cols;i++){
		uint32	c=pal[i];
		*((uchar*)pal+i*3)=c>>16;		//g
		*((uchar*)pal+i*3+1)=c>>8;		//r
		*((uchar*)pal+i*3+2)=c>>24;		//b
	}
	stt=output->Write(pal,len=3*cols);			//  write palette
	if(stt<len)return(stt);
	stt=output->Write(fa,len=fasize);			//  write flaga
	if(stt<len)return(stt);
	stt=output->Write(fb,len=fbsize);			//  write flagb
	if(stt<len)return(stt);
	stt=output->Write(fc,len=fcsize);			//  write pixel
	if(stt<len)return(stt);
	return(B_OK);
}

static status_t ReadMAKI(BPositionIO *input,BPositionIO *output,BMessage *ioExtension)
{																//MAKI->BITMAP
	MAKIChecker			makick;
	TranslatorBitmap	bmphd;
	uint32				pal[256];
	char		mode;
	ssize_t		len,stt;

	//	read MAKI checker
	stt=input->Read(&makick,len=sizeof(makick));
	if(stt<len)return(stt);
	if(!strncmp(makick.mark,"MAKI01A ",8))mode=1;
	else if(!strncmp(makick.mark,"MAKI01B ",8))mode=2;
	else if(!strncmp(makick.mark,"MAKI02  ",8))mode=3;
	else return(B_NO_TRANSLATOR);

	if((mode==1)||(mode==2)){			//MAKI
		MAKIHeader		makihd;
		//  read MAKI header
		stt=input->Read(&makihd,len=sizeof(makihd));
		if(stt<len)return(stt);
		makihd.fb_size=B_BENDIAN_TO_HOST_INT16(makihd.fb_size);
		makihd.pa_size=B_BENDIAN_TO_HOST_INT16(makihd.pa_size);
		makihd.pb_size=B_BENDIAN_TO_HOST_INT16(makihd.pb_size);
		makihd.exflag=B_BENDIAN_TO_HOST_INT16(makihd.exflag);
		makihd.xmin=B_BENDIAN_TO_HOST_INT16(makihd.xmin);
		makihd.ymin=B_BENDIAN_TO_HOST_INT16(makihd.ymin);
		makihd.xmax=B_BENDIAN_TO_HOST_INT16(makihd.xmax);
		makihd.ymax=B_BENDIAN_TO_HOST_INT16(makihd.ymax);

		if((rmode>0)&&(makihd.exflag&1)){makihd.ymin*=2;makihd.ymax*=2;}			//for aspect(1:2)

		//	write bitmap header(output=32bit only)
		bmphd.rowBytes=makihd.xmax*4;
		bmphd.dataSize=bmphd.rowBytes*makihd.ymax;
		stt=output->SetSize(sizeof(bmphd)+bmphd.dataSize);
		if(stt<B_OK)return(stt);
		bmphd.magic=B_HOST_TO_BENDIAN_INT32(B_TRANSLATOR_BITMAP);
		bmphd.bounds.left=B_HOST_TO_BENDIAN_FLOAT((float)makihd.xmin);
		bmphd.bounds.top=B_HOST_TO_BENDIAN_FLOAT((float)makihd.ymin);
		bmphd.bounds.right=B_HOST_TO_BENDIAN_FLOAT((float)makihd.xmax-1);
		bmphd.bounds.bottom=B_HOST_TO_BENDIAN_FLOAT((float)makihd.ymax-1);
//		bmphd.colors=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		*(int32*)(&(bmphd.colors))=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		bmphd.rowBytes=B_HOST_TO_BENDIAN_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_HOST_TO_BENDIAN_INT32(bmphd.dataSize);
		stt=output->Write(&bmphd,len=sizeof(bmphd));
		if(stt<len)return(stt);
		bmphd.magic=B_BENDIAN_TO_HOST_INT32(bmphd.magic);
		bmphd.bounds.left=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.left);
		bmphd.bounds.top=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.top);
		bmphd.bounds.right=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.right);
		bmphd.bounds.bottom=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.bottom);
//		bmphd.colors=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		*(int32*)(&(bmphd.colors))=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		bmphd.rowBytes=B_BENDIAN_TO_HOST_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_BENDIAN_TO_HOST_INT32(bmphd.dataSize);

		if((rmode>0)&&(makihd.exflag&1)){makihd.ymin/=2;makihd.ymax/=2;}			//for aspect(1:2)

		//	xfer data
		int	w=makihd.xmax-makihd.xmin;
		int	h=makihd.ymax-makihd.ymin;
		uint	wbsize=w*h*4;
		wb=new uchar[wbsize];		if(!wb)return(B_NO_MEMORY);
		memset(wb,0,wbsize);					//clr output-area
		stt=input->Read(pal,len=3*16);			//  read palette
		if(stt<len)return(stt);
		for(int i=15;i>=0;i--){
			uchar	g=*((uchar*)pal+i*3);
			uchar	r=*((uchar*)pal+i*3+1);
			uchar	b=*((uchar*)pal+i*3+2);
			pal[i]=B_HOST_TO_BENDIAN_INT32((((uint32)b)<<24)|(((uint32)g)<<16)|(((uint32)r)<<8));
		}
		uint	fasize=w*h/256;
		uint	fbsize=makihd.fb_size;
		uint	fcsize=makihd.pa_size+makihd.pb_size;
		uchar	*fap,*fa=wb;			//flaga(read)
		uchar	*fbp,*fb=fa+fasize;		//flagb(read)
		uchar	*fcp,*fc=fb+fbsize;		//pixel(read)
		uchar	*fdp,*fd=wb+wbsize/2;	//work
		uchar	*ffp,*ff=wb+wbsize-w*h;	//color
		uint32	*fgp,*fg=(uint32*)wb;
		stt=input->Read(fa,len=fasize+fbsize+fcsize);	//  read all
		if(stt<len)return(stt);

		fap=fa;fbp=fb;fdp=fd;
		uchar	f,b=0;
		int		dy=w/16;
		for(int y=0;y<h;y+=4){
			for(int x=0;x<w;x+=16){
				if(b>1){
					b>>=1;
				}else{
					b=0x80;
					f=*(fap++);
				}
				uchar	f1,f2,f3,f4;
				f1=f2=f3=f4=0;
				if(f&b){
					f1=(*fbp)&0xf0;
					f2=((*fbp)&0x0f)<<4;
					fbp++;
					f3=(*fbp)&0xf0;
					f4=((*fbp)&0x0f)<<4;
					fbp++;
				}
				b>>=1;
				if(f&b){
					f1|=(*fbp)>>4;
					f2|=(*fbp)&0x0f;
					fbp++;
					f3|=(*fbp)>>4;
					f4|=(*fbp)&0x0f;
					fbp++;
				}
				(*fdp)=f1;	fdp+=dy;
				(*fdp)=f2;	fdp+=dy;
				(*fdp)=f3;	fdp+=dy;
				(*fdp)=f4;	fdp=fdp-3*dy+1;
			}
			fdp+=3*dy;
		}
		fcp=fc;fdp=fd;ffp=ff;
		uchar	*fep=(mode==1)?(ffp-2*w):(ffp-4*w);
		b=0;
		for(int y=0;y<h;y++)for(int x=0;x<w;x+=2){
			if(b>1){
				b>>=1;
			}else{
				b=0x80;
				f=*(fdp++);
			}
			if(f&b){
				*(ffp++)=(fep<ff)?(*fcp)>>4:(*fep)^((*fcp)>>4);
				fep++;
				*(ffp++)=(fep<ff)?(*fcp)&0x0f:(*fep)^((*fcp)&0x0f);
				fep++;fcp++;
			}else{
				*(ffp++)=(fep<ff)?0:(*fep);
				fep++;
				*(ffp++)=(fep<ff)?0:(*fep);
				fep++;
			}
		}
		ffp=ff;fgp=fg;
		for(int y=0;y<h;y++)for(int x=0;x<w;x++){
			*(fgp++)=pal[*(ffp++)];
		}
		fgp=fg;
		for(int y=0;y<h;y++){
			stt=output->Write(fgp,len=bmphd.rowBytes);
			if(stt<len)return(stt);
			if((rmode>0)&&(makihd.exflag&1)){										//for aspect(1:2)
				stt=output->Write(fgp,len=bmphd.rowBytes);
				if(stt<len)return(stt);
			}
			fgp+=w;
		}
	}else{								//MAG
		MAGHeader		maghd;
		//  read MAG header
		if(makick.user[19]!=0x1a)while(1){
			stt=input->Read(&maghd,1);
			if(stt<1)return(stt);
			if(maghd.top==0x1a)break;
		}
		stt=input->Read(&maghd,len=sizeof(maghd));
		if(stt<len)return(stt);
		maghd.xmin=B_LENDIAN_TO_HOST_INT16(maghd.xmin);
		maghd.ymin=B_LENDIAN_TO_HOST_INT16(maghd.ymin);
		maghd.xmax=B_LENDIAN_TO_HOST_INT16(maghd.xmax);
		maghd.ymax=B_LENDIAN_TO_HOST_INT16(maghd.ymax);
		maghd.fa_off=B_LENDIAN_TO_HOST_INT32(maghd.fa_off);
		maghd.fb_off=B_LENDIAN_TO_HOST_INT32(maghd.fb_off);
		maghd.fb_size=B_LENDIAN_TO_HOST_INT32(maghd.fb_size);
		maghd.p_off=B_LENDIAN_TO_HOST_INT32(maghd.p_off);
		maghd.p_size=B_LENDIAN_TO_HOST_INT32(maghd.p_size);

		if((rmode>0)&&(maghd.scrn&1)){maghd.ymin*=2;maghd.ymax=maghd.ymax*2+1;}		//for aspect(1:2)

		//	write bitmap header(output=32bit only)
		bmphd.rowBytes=(maghd.xmax-maghd.xmin+1)*4;
		bmphd.dataSize=bmphd.rowBytes*(maghd.ymax-maghd.ymin+1);
		stt=output->SetSize(sizeof(bmphd)+bmphd.dataSize);
		if(stt<B_OK)return(stt);
		bmphd.magic=B_HOST_TO_BENDIAN_INT32(B_TRANSLATOR_BITMAP);
		bmphd.bounds.left=B_HOST_TO_BENDIAN_FLOAT((float)maghd.xmin);
		bmphd.bounds.top=B_HOST_TO_BENDIAN_FLOAT((float)maghd.ymin);
		bmphd.bounds.right=B_HOST_TO_BENDIAN_FLOAT((float)maghd.xmax);
		bmphd.bounds.bottom=B_HOST_TO_BENDIAN_FLOAT((float)maghd.ymax);
//		bmphd.colors=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		*(int32*)(&(bmphd.colors))=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		bmphd.rowBytes=B_HOST_TO_BENDIAN_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_HOST_TO_BENDIAN_INT32(bmphd.dataSize);
		stt=output->Write(&bmphd,len=sizeof(bmphd));
		if(stt<len)return(stt);
		bmphd.magic=B_BENDIAN_TO_HOST_INT32(bmphd.magic);
		bmphd.bounds.left=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.left);
		bmphd.bounds.top=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.top);
		bmphd.bounds.right=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.right);
		bmphd.bounds.bottom=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.bottom);
//		bmphd.colors=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		*(int32*)(&(bmphd.colors))=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		bmphd.rowBytes=B_BENDIAN_TO_HOST_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_BENDIAN_TO_HOST_INT32(bmphd.dataSize);

		if((rmode>0)&&(maghd.scrn&1)){maghd.ymin/=2;maghd.ymax/=2;}					//for aspect(1:2)
		if(maghd.scrn&0x80){														//for odd_pixels
			maghd.xmin&=0xfffc;maghd.xmax|=0x0003;
		}else{
			maghd.xmin&=0xfff8;maghd.xmax|=0x0007;
		}

		//	xfer data
		int	w=maghd.xmax-maghd.xmin+1;
		int	h=maghd.ymax-maghd.ymin+1;
		uint	wbsize=w*h*4;
		wb=new uchar[wbsize];		if(!wb)return(B_NO_MEMORY);
		memset(wb,0,wbsize);					//clr output-area
		int	cols=(maghd.scrn&0x80)?256:16;
		stt=input->Read(pal,len=3*cols);		//  read palette
		if(stt<len)return(stt);
		for(int i=cols-1;i>=0;i--){
			uchar	g=*((uchar*)pal+i*3);
			uchar	r=*((uchar*)pal+i*3+1);
			uchar	b=*((uchar*)pal+i*3+2);
			pal[i]=B_HOST_TO_BENDIAN_INT32((((uint32)b)<<24)|(((uint32)g)<<16)|(((uint32)r)<<8));
		}
		uint	fasize=maghd.fb_off-maghd.fa_off;
		uint	fbsize=maghd.p_off-maghd.fb_off;
		uint	fcsize=maghd.p_size;
		uchar	*fap,*fa=wb;			//flaga(read)
		uchar	*fbp,*fb=fa+fasize;		//flagb(read)
		uchar	*fcp,*fc=fb+fbsize;		//pixel(read)
		uchar	*fdp,*fd=wb+wbsize/2;	//copy-position
		uchar	*fep,*fe=wb+wbsize*5/8;	//work
		uchar	*ffp,*ff=wb+wbsize-w*h;	//color
		uint32	*fgp,*fg=(uint32*)wb;
		stt=input->Read(fa,len=fasize+fbsize+fcsize);	//  read all
		if(stt<len)return(stt);
		int		xd=(maghd.scrn&0x80)?2:4;

		fap=fa;fbp=fb;fdp=fd;
		uchar	f,b=0;
		for(int y=0;y<h;y++){
			fep=fe;
			for(int x=0;x<w;x+=xd*2){
				if(b>1){
					b>>=1;
				}else{
					b=0x80;
					f=*(fap++);
				}
				if(f&b)(*fep)^=*(fbp++);
				*(fdp++)=(*fep)>>4;
				*(fdp++)=(*fep)&0x0f;
				fep++;
			}
		}
		int	d[16];	for(int i=0;i<16;i++)d[i]=w*mag_dy[i]+xd*mag_dx[i];
		fcp=fc;fdp=fd;ffp=ff;
		for(int y=0;y<h;y++)for(int x=0;x<w;x+=xd){
			uchar	dd=*(fdp++);
			if(dd){
				memcpy(ffp,ffp+d[dd],xd);
				ffp+=xd;
			}else{
				if(cols==256){			//256color
					*(ffp++)=*(fcp++);
					*(ffp++)=*(fcp++);
				}else{					//16color
					*(ffp++)=(*fcp)>>4;
					*(ffp++)=(*fcp)&0x0f;
					fcp++;
					*(ffp++)=(*fcp)>>4;
					*(ffp++)=(*fcp)&0x0f;
					fcp++;
				}
			}
		}
		ffp=ff;fgp=fg;
		for(int y=0;y<h;y++)for(int x=0;x<w;x++){
			*(fgp++)=pal[*(ffp++)];
		}
		fgp=fg+(int)(bmphd.bounds.left-maghd.xmin);
		for(int y=0;y<h;y++){
			stt=output->Write(fgp,len=bmphd.rowBytes);
			if(stt<len)return(stt);
			if((rmode>0)&&(maghd.scrn&1)){											//for aspect(1:2)
				stt=output->Write(fgp,len=bmphd.rowBytes);
				if(stt<len)return(stt);
			}
			fgp+=w;
		}
	}
	return(B_OK);
}

status_t Translate(BPositionIO *inSource,const translator_info *inInfo,BMessage *ioExtension,uint32 outType,BPositionIO *outDestination)
{
	status_t	stt=B_NO_TRANSLATOR;
	rb=wb=NULL;
	if(outType==inInfo->type)stt=CopyLoop(inSource,outDestination);
	if((outType=='MAG ')&&(inInfo->type==B_TRANSLATOR_BITMAP))stt=WriteMAKI(inSource,outDestination,ioExtension);
	if((outType==B_TRANSLATOR_BITMAP)&&(inInfo->type=='MAKI'))stt=ReadMAKI(inSource,outDestination,ioExtension);
	if((outType==B_TRANSLATOR_BITMAP)&&(inInfo->type=='MAG '))stt=ReadMAKI(inSource,outDestination,ioExtension);
	delete rb;delete wb;
	return(stt);
}

status_t MakeConfig(BMessage *ioExtension,BView **outView,BRect *outExtent)
{
	*outView=new configview(BRect(0,0,200,150));
	outExtent->Set(0,0,200,150);
	return(B_OK);
}

status_t GetConfigMessage(BMessage *ioExtension)
{
	return(B_OK);
}
